"""
=====================================================
Finite elements based on the direct stiffness method
=====================================================

Elements in this group share the following features:
* small deformation kinematics
* linear elastic material behavior
* based on the analytical solution of the governing equations
"""

__all__ = (
    "Truss",
    "Beam2D",
    "Frame2D",
)

from .Beam2D import *
from .Frame2D import *
from .Truss import *
